/*  mingetty.c
 *
 *  Copyright (C) 1996 Florian La Roche
 *  florian@jurix.jura.uni-sb.de florian@suse.de florian@knorke.saar.de
 *  Copyright (C) 1999-2010 Werner Fink <werner@suse.de>
 *  Copyright (C) 1999-2001 SuSE Linux GmbH
 *  Copyright (C) 2002 SuSE Linux AG
 *  Copyright (C) 2010 SUSE LINUX Products GmbH
 *
 *  Newer versions should be on ftp://ftp.suse.com/pub/projects/init/
 *
 *  utmp-handling is from agetty in util-linux 2.5 (probably from
 *  Peter Orbaek <poe@daimi.aau.dk>)
 *
 *  SuSE - GmbH has paid some of the time that was needed to write
 *  this program. Thanks a lot for their support.
 *
 *  This getty can only be used as console getty. It is very small, but
 *  should be very reliable. For a modem getty, I'd also use nothing else
 *  but mgetty.
 *
 *  Usage: mingetty [--noclear] [--noreset] [--long-hostname]
 *	 [--no-hostname] [--login program] [--old]
 *	 [--logopts "options for login prog"] tty [term]
 *  Example entry in /etc/inittab: 1:123:respawn:/sbin/mingetty tty1
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version
 *  2 of the License, or (at your option) any later version.
 *
 *  Kurt Garloff <garloff@suse.de>, Feb 98, Jun 99:
 *  Added --login and --logopts options to let the superuser choose
 *    the login program and its options used. 
 *    \\u will be replaced by the user name.
 *  Example (/etc/inittab):
 *  13:23:respawn:/sbin/mingetty --login /usr/bin/ssh --logopts "\\u@target" tty13
 *  mingetty then execs "/usr/bin/ssh logname@target" 
 *  instead of          "/bin/login -- logname"
 * SECURITY notice: mingetty checks the name for a leading - to make sure
 *  the user can not easily pass an option instead of a logname.
 * Better: If the program supports it, pass -- before the username gets
 *  passed, which will prevent the interpretation as options.
 *
 *  Tue Sep  7 21:20:01 CEST 1999 - Werner Fink <werner@suse.de>
 *	Enable and revisit resetting of tty.
 *	Use wtmp interface of glibc
 *	Set tty process group
 *	Restore signal handling in signal handler
 *
 *  Wed May 10 23:25:02 CEST 2000 - Kurt Garloff <garloff@suse.de>
 *	Add support for devfs by delaying the /dev/vcs[a] 
 *	permission changes.
 *
 *  Mon Oct 23 12:40:00 CEST 2000 - Werner Fink <werner@suse.de>
 *	On ro files systems chmod/chown shouldn't force an exit.
 *	Therefore we need a warn message.
 *	Due to the lack of a revoke system call we signal
 *	all file holders including the mmaping processes.
 *
 *  Wed Apr 24 18:54:40 CEST 2002 - Werner Fink <werner@suse.de>
 *	Add support for other TTY devices than virtual console
 *	Add nohost patch from Anders ??
 *
 *  Wed Nov 13 18:09:12 CET 2002 - Werner Fink <werner@suse.de>
 *	Add an old option for avoiding vcs/vcsa hangup, terminal
 *	reset, and the usage of the glibc for updating wtmp.
 *	Cleanup the source and the manual page.
 *
 *  Tue Mar 30 13:53:48 CEST 2004 - werner@suse.de
 *	More about UTF-8: enable the kernel to erease UTF-8 multi byte
 *	characters. Therefore add workaround if glibc bits/termios.h
 *	is missing the IUTF8 input flag.
 *
 *  Tue May 29 16:55:52 CEST 2007 - Werner Fink <werner@suse.de>
 *	Adopt autologin and other new feature of the other mingetty
 *	at sourceforge.
 *
 *  Fri Jul 23 13:24:00 CEST 2010 - Werner Fink <werner@suse.de>
 *	Adopt loginpause feature of the other mingetty at sourceforge.
 *      Ignore SIGQUIT and SIGINT and switch Numlock on.
 */

#ifndef VERSION
# define VERSION "mingetty-1.0.8s"
#endif

#ifndef DEFAULT_TERM
# define DEFAULT_TERM	"linux"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <termios.h>
#include <string.h>
#include <sys/ioctl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <signal.h>
#include <fcntl.h>
#include <stdarg.h>
#include <ctype.h>
#include <utmp.h>
#include <getopt.h>
#include <sys/param.h>
#include <syslog.h>
#include <sys/utsname.h>
#include <time.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <grp.h>
#include <locale.h>
#include <iconv.h>
#include <wctype.h>
#include <sys/kd.h>
#include "revoke.h"

#ifndef IUTF8
#  error IUTF8 input flag not defined
#endif

#define	ISSUE "/etc/issue"	/* displayed before the login prompt */
#define LOGIN "login: "		/* login prompt */
#ifndef _PATH_LOGIN
#define _PATH_LOGIN "/bin/login"
#endif

/* name of this program (argv[0]) */
static char *progname;
/* on which tty line are we sitting? (e.g. tty1) */
static char *tty;
static char *vcline = NULL;
/* some information about this host */
static struct utsname uts;
/* the hostname */
static char hn[MAXHOSTNAMELEN + 1];
/* process and session ID of this program */
static pid_t pid, sid;
/* login program invoked */
static char *loginprog = _PATH_LOGIN;
/* Do not send a reset string to the terminal. */
static int noclear = 0;
/* Do not print a newline. */
static int nonewline = 0;
/* Do not print /etc/issue. */
static int noissue = 0;
/* Do not call vhangup() on the tty. */
static int nohangup = 0;
/* Do not print any hostname. */
static int nohostname = 0;
/* Print the whole string of gethostname() instead of just until the next "." */
static int longhostname = 0;
/* time to wait, seconds */
static int delay = 0;
/* chroot directory */
static char *ch_root = NULL;
/* working directory to change into */
static char *ch_dir = NULL;
/* 'nice' level of the program */
static int priority = 0;
/* automatic login with this user */
static char *autologin = NULL;
/* try to read a char before dropping to login prompt */
static int loginpause = 0;
/* do not reset the terminal ? */
static int noreset = 0;
/* current time */
static int old = 0;
/* current time */
static int mode = K_RAW;

/*
 * output warn/error messages
 */
static void _vlogger (int priority, const char *fmt, va_list va_alist)
{
	char buf[256] =  {'\0'}, *bp = &buf[0];

	strncpy (buf, progname, sizeof(buf) - 4);
	buf[sizeof(buf) - 5] = '\0';
	strcat (buf, ": ");
	bp = buf + strlen (buf);
	vsnprintf (bp, sizeof(buf)-strlen(buf), fmt, va_alist);
	openlog (progname, LOG_PID, LOG_AUTH);
	syslog (priority, "%s", buf);
	closelog ();
}

/*
 * output error messages
 */
void error (const char *fmt, ...)
{
	va_list va_alist;
	va_start (va_alist, fmt);
	_vlogger (LOG_ERR, fmt, va_alist);
	va_end (va_alist);
	exit (1);
}

/*
 * output warn messages
 */
void warn (const char *fmt, ...)
{
	va_list va_alist;
	va_start (va_alist, fmt);
	_vlogger (LOG_WARNING, fmt, va_alist);
	va_end (va_alist);
	return;
}

/*
 * output alert messages
 */
static __inline__ void alert (const char *fmt, ...)
{
	va_list va_alist;
	va_start (va_alist, fmt);
	_vlogger (LOG_ALERT, fmt, va_alist);
	va_end (va_alist);
	return;
}

/*
 * update_utmp - update our utmp entry
 *
 * The utmp file holds miscellaneous information about things started by
 * /sbin/init and other system-related events. Our purpose is to update
 * the utmp entry for the current process, in particular the process
 * type and the tty line we are listening to. Return successfully only
 * if the utmp file can be opened for update, and if we are able to find
 * our entry in the utmp file.
 */
static void update_utmp (void)
{
	struct utmp ut;
	struct utmp *utp;
	time_t cur_time;

	utmpname (_PATH_UTMP);
	setutent ();
	while ((utp = getutent ()))
		if (utp->ut_type == INIT_PROCESS && utp->ut_pid == pid)
			break;

	if (utp) {
		memcpy (&ut, utp, sizeof (ut));
	} else {
		/* some inits don't initialize utmp... */
		memset (&ut, 0, sizeof (ut));
		if (vcline)
			/* Standard virtual console devices */
			strncpy (ut.ut_id, vcline, sizeof (ut.ut_id));
		else {
			size_t len = strlen(tty);
			char * ptr;
			if (len >= sizeof (ut.ut_id))
				ptr = tty + len - sizeof (ut.ut_id);
			else
				ptr = tty;
			strncpy (ut.ut_id, ptr, sizeof (ut.ut_id));
		}
	}

	strncpy (ut.ut_user, "LOGIN", sizeof (ut.ut_user));
	strncpy (ut.ut_line, tty, sizeof (ut.ut_line));

	time (&cur_time);
	ut.ut_time = cur_time;
	ut.ut_type = LOGIN_PROCESS;
	ut.ut_pid = pid;
	ut.ut_session = sid;

	pututline (&ut);
	endutent ();

	updwtmp (_PATH_WTMP, &ut);
}

/* open_tty - set up tty as standard { input, output, error } */
static void open_tty (void)
{
	char buf[PATH_MAX+1];
	int fd, len;
	pid_t tid;
	gid_t gid = 0;
	struct group *gr = NULL;

	/* Use tty group if available */
	if ((gr = getgrnam("tty")))
		gid = gr->gr_gid;

	/* Set up new standard input. */
	if (((len = snprintf(buf, sizeof(buf), "/dev/%s", tty)) >= sizeof(buf)) || (len < 0))
		error ("%s: %s", buf, strerror(EOVERFLOW));

	/* There is always a race between this reset and the call to
	   vhangup() that s.o. can use to get access to your tty. */
	if (chown (buf, 0, gid) || chmod (buf, (gid ? 0660 : 0600))) {
		if (errno == EROFS)
			warn("%s: %s", buf, strerror(errno));
		else
			error ("%s: %s", buf, strerror(errno));
	}

	if ((fd = open (buf, O_RDWR|O_NOCTTY, 0)) < 0)
		error ("%s: cannot open tty: %s", buf, strerror(errno));
	if (!isatty (fd))
		error ("%s: not a tty", buf);
	if (((tid = tcgetsid(fd)) < 0) || (pid != tid)) {
		if (ioctl (fd, TIOCSCTTY, 1) == -1)
			error ("%s: cannot get controlling tty: %s", buf, strerror(errno));
	}

	if (nohangup == 0) {
		/* vhangup() will replace all open file descriptors in the kernel
		   that point to our controlling tty by a dummy that will deny
		   further reading/writing to our device. It will also reset the
		   tty to sane defaults, so we don't have to modify the tty device
		   for sane settings. We also get a SIGHUP/SIGCONT.
		 */
		if (vhangup ())
			error ("%s: vhangup() failed", tty);
		(void)ioctl (fd, TIOCNOTTY);
	}

	/* Get rid of the present stdout/stderr. */
	close (2);
	close (1);
	close (0);
	if (fd > 2)
		close (fd);

	if ((fd = open (buf, O_RDWR|O_NOCTTY, 0)) < 0)
		error ("%s: cannot open tty: %s", buf, strerror(errno));
	if (((tid = tcgetsid(fd)) < 0) || (pid != tid)) {
		if (ioctl (fd, TIOCSCTTY, 1) == -1)
			error ("%s: cannot get controlling tty: %s", buf, strerror(errno));
	}
	if (tcsetpgrp(fd, pid))
		error ("%s: cannot set process group: %s", buf, strerror(errno));

	/* Set up standard output and standard error file descriptors. */
	if (dup2 (fd, 0) != 0)
		error ("%s: dup problem: %s", buf, strerror(errno));
	if (dup2 (fd, 1) != 1)
		error ("%s: dup problem: %s", buf, strerror(errno));
	if (dup2 (fd, 2) != 2)
		error ("%s: dup problem: %s", buf, strerror(errno));
	if (fd > 2)
		close (fd);

#ifdef IUTF8
	/* Detect mode of current keyboard setup, e.g. for UTF-8 */
	if (ioctl(0, KDGKBMODE, &mode) < 0)
		mode = K_RAW;
#endif

	/* Write a reset string to the terminal. This is very linux-specific
	   and should be checked for other systems. */
	if (noclear == 0)
		/* don't write a full reset (ESC c) because this leaves the 
		   unicode mode again if the terminal was in unicode mode
		   and also undos the ESC sequences in CONSOLE_MAGIC which
		   are needed for some languages/console-fonts.
		   Just put the cursor to the home position (ESC [ H),
		   erase everything below the cursor (ESC [ J), and set the
		   scrolling region to the full window (ESC [ r) */
		write (0, "\033[r\033[H\033[J", 9);

	/*
	 * Only for the standard virtual console devices
	 */
	if (!old && vcline)
		vcshangup(gid, vcline);

	/*
	 * The above reset only puts the output things into a sane state.
	 * The input state is not reset.
	 */
	if (!noreset && !old)
		reset_stdin(mode);
}

static void output_special_char (unsigned char c)
{
	switch (c) {
	case 's':
		printf ("%s", uts.sysname);
		break;
	case 'n':
		printf ("%s", uts.nodename);
		break;
	case 'r':
		printf ("%s", uts.release);
		break;
	case 'v':
		printf ("%s", uts.version);
		break;
	case 'm':
		printf ("%s", uts.machine);
		break;
	case 'o':
		printf ("%s", uts.domainname);
		break;
	case 'd':
	case 't':
		{
			time_t cur_time;
			struct tm *tm;

			time (&cur_time);
			tm = localtime (&cur_time);
			if (c == 'd') /* ISO 8601 */
				printf ("%d-%02d-%02d", 1900 + tm->tm_year,
					tm->tm_mon + 1, tm->tm_mday);
			else
				printf ("%02d:%02d:%02d", tm->tm_hour,
					tm->tm_min, tm->tm_sec);
			break;
		}
	case 'l':
		printf ("%s", tty);
		break;
	case 'u':
	case 'U':
		{
			int users = 0;
			struct utmp *ut;
			setutent ();
			while ((ut = getutent ()))
				if (ut->ut_type == USER_PROCESS)
					users++;
			endutent ();
			printf ("%d", users);
			if (c == 'U')
				printf (" user%s", users == 1 ? "" : "s");
			break;
		}
	default:
		putchar (c);
	}
}

/* do_prompt - show login prompt, optionally preceded by /etc/issue contents */
static void do_prompt (int showlogin)
{
	FILE *fd;
	int c;

	if (nonewline == 0)
		putchar ('\n');
	if (noissue == 0 && (fd = fopen (ISSUE, "r"))) {
		while ((c = getc (fd)) != EOF) {
			if (c == '\\')
				output_special_char (getc (fd));
			else
				putchar (c);
		}
		fclose (fd);
	}
	if (loginpause) {
		puts ("[press ENTER to login]");
		getc (stdin);
	}
	if (showlogin) {
		int kb = 0, nl = 0;
		struct stat st;
		if (stat("/var/run/numlock-on", &st) == 0)
			nl = 1;
		if (ioctl(0, KDGKBLED, &kb) == 0) {
			char warn[128];
			off_t len = 0;

			if (nl && (kb & 0x02) == 0) {
				strcpy(&warn[0], "Num Lock off");
				len += 12;
			} else if (nl == 0 && (kb & 2) && (kb & 0x20) == 0) {
				strcpy(&warn[0], "Num Lock on");
				len += 11;
			}

			if ((kb & 0x04) && (kb & 0x40) == 0) {
				if (len) {
					strcpy(&warn[len], ", ");
					len += 2;
				}
				strcpy(&warn[len], "Caps Lock on");
				len += 12;
			}

			if ((kb & 0x01) && (kb & 0x10) == 0) {
				if (len) {
					strcpy(&warn[len], ", ");
					len += 2;
				}
				strcpy(&warn[len], "Scroll Lock on");
				len += 14;
			}

			if (len)
				printf ("Hint: %s\n\n", warn);
		}
	}
	if (nohostname == 0)
		printf ("%s ", hn);
	if (showlogin)
		printf ("%s", LOGIN);
	fflush (stdout);
}

/* get_logname - get user name, establish speed, erase, kill, eol */
static char *get_logname (void)
{
	static char logname[4*UT_NAMESIZE];
	char *bp;
	unsigned char c;
	int ascii;
	iconv_t ic;

	tcflush (0, TCIFLUSH);		/* flush pending input */

#ifdef IUTF8
	/* Check for UTF-8 mode */
	switch(mode) {
		case K_UNICODE:
			ascii = 0;
			setlocale(LC_CTYPE, "en_US.UTF-8");
			break;
		case K_RAW:
		case K_MEDIUMRAW:
		case K_XLATE:
		default:
			ascii = 1;
			setlocale(LC_CTYPE, "POSIX");
			break;
	}
#endif

	for (*logname = 0; *logname == 0;) {
		do_prompt (1);
		for (bp = logname;;) {
			if (read (0, &c, 1) < 1) {
				if (errno == EINTR || errno == EAGAIN) {
					usleep(1000);
					continue;
				}
				if (errno == EIO || errno == ENOENT)
					exit (EXIT_SUCCESS);
				error ("%s: read: %s", tty, strerror (errno));
			}

			if (c == '\n' || c == '\r') {
				*bp = 0;
				break;
			}

			if (ascii && !isprint (c))
				error ("%s: invalid character 0x%x in login name", tty, c);

			if ((size_t)(bp - logname) >= sizeof (logname) - 1)
				error ("%s: too long login name", tty);

			*bp++ = c;
		}
	}

	if (!ascii && (ic = iconv_open("WCHAR_T", "UTF-8"))) {
		char tmpbuf[4*sizeof(logname)], *op, *lp;
		size_t len = bp - logname;
		size_t out = sizeof(tmpbuf) - 1;
		size_t wcl;
		wint_t *wcp;

		op = tmpbuf;
		lp = logname;
		if ((wcl = iconv(ic , &lp, &len, &op, &out)) == (size_t)-1)
			error ("%s: invalid character conversion for login name", tty);
		iconv_close(ic);

		wcp = (wint_t*)tmpbuf;
		wcp[wcl] = (wint_t)0;
		while (*wcp) {
			const wint_t wc = *wcp++;
			if (!iswprint(wc))
				error ("%s: invalid character 0x%x in login name", tty, wc);
		}
	}
	return logname;
}

static void usage (void)
{
	error ("%s: usage: '%s [--noclear] [--nonewline] [--noissue]"
	       " [--nohangup] [--noreset] [--no-hostname] [--long-hostname]"
	       " [--login program] [--logopts \"loginprg opts\"] [--nice 10]"
	       " [--delay 10] [--chdir /home] [--chroot /chroot] [--autologin user]"
	       " [--loginpause] [--old] tty [term]' with e.g. tty=tty1", VERSION, progname);
}

static struct option const long_options[] = {
	{ "autologin", required_argument, NULL, 'a' },
        { "loginpause", no_argument, &loginpause, 1 },
	{ "chdir", required_argument, NULL, 'w' },
	{ "chroot", required_argument, NULL, 'r' },
	{ "delay", required_argument, NULL, 'd' },
	{ "noclear", no_argument, &noclear, 1 },
	{ "nonewline", no_argument, &nonewline, 1 },
	{ "noissue", no_argument, &noissue, 1 },
	{ "nohangup", no_argument, &nohangup, 1 },
	{ "noreset", no_argument, &noreset, 1 },
	{ "no-hostname", no_argument, &nohostname, 1 },
	{ "nohostname", no_argument, &nohostname, 1 },	/* compat option */
	{ "login", required_argument, NULL, 'l' },
	{ "loginprog", required_argument, NULL, 'l' },	/* compat option */
	{ "loginopts", required_argument, NULL, 'o' },
	{ "logopts", required_argument, NULL, 'o' },
	{ "long-hostname", no_argument, &longhostname, 1 },
	{ "nice", required_argument, NULL, 'n' },
	{ "old", no_argument, &old, 1 },
	{ 0, 0, 0, 0 }
};

/*
 * main program
 */
int main (int argc, char **argv)
{
	sigset_t set;
	char *logname, *s;
	int c;
	char *logoptstr;
	char logcmd[NAME_MAX+1];
	char *logarr[ARRAY_SIZE_MAX];
	logoptstr = "-- \\u";
	struct sigaction sa, sa_hup, sa_quit, sa_int;

	sa.sa_handler = SIG_IGN;
	sa.sa_flags = SA_RESTART;
	sigemptyset (&sa.sa_mask);
	sigaction (SIGHUP, &sa, &sa_hup);
	sigaction (SIGQUIT, &sa, &sa_quit);
	sigaction (SIGINT, &sa, &sa_int);

	progname = basename(argv[0]);
	if (!progname)
		progname = "mingetty";

	uname (&uts);
	gethostname (hn, MAXHOSTNAMELEN);
	hn[MAXHOSTNAMELEN] = '\0';
	pid = getpid ();
	sid = getsid (0);

	while ((c = getopt_long (argc, argv, "a:d:l:n:w:r:", long_options, (int *) 0))
			!= EOF) {
		switch (c) {
		case 0:
			break;
		case 'a':
			if (!optarg) usage ();
			autologin = optarg;
			break;
		case 'd':
			if (!optarg) usage ();
			delay = atoi (optarg);
			break;
		case 'l':
			if (!optarg) usage ();
			loginprog = optarg;
			break;
		case 'o':
			if (!optarg) usage ();
			logoptstr = optarg;
			break;
		case 'n':
			if (!optarg) usage ();
			priority = atoi (optarg);
			break;
		case 'r':
			if (!optarg) usage ();
			ch_root = optarg;
			break;
		case 'w':
			if (!optarg) usage ();
			ch_dir = optarg;
			break;
		default:
			usage ();
		}
	}
	if (longhostname == 0 && (s = strchr (hn, '.')))
		*s = '\0';
	tty = argv[optind++];
	if (!tty)
		usage ();

	/* Skip the "/dev/", we add it later */
	if (strncmp (tty, "/dev/", 5) == 0)
		tty += 5;

	/* On virtual console remember the line which is used for */
	if (strncmp(tty, "tty", 3) == 0 && strspn(tty + 3, "0123456789") == strlen(tty+3))
		vcline = tty+3;

	/* Set terminal type if given as last argument */
	if (argc > optind && argv[optind])
		setenv("TERM", argv[optind], 1);
	else
		setenv("TERM", DEFAULT_TERM, 1);

	update_utmp ();
	if (delay)
		sleep (delay);
	open_tty ();

	/* flush input and output queues, important for modems */
	tcflush(0, TCIOFLUSH);


	/* unmask SIGHUP if inherited */
	sigemptyset (&set);
	sigaddset (&set, SIGHUP);
	sigprocmask (SIG_UNBLOCK, &set, NULL);
	sigaction (SIGHUP, &sa_hup, NULL);

	if (autologin) {
		do_prompt (0);
		printf ("%s%s (automatic login)\n", LOGIN, autologin);
		logname = autologin;
		logoptstr = "-f \\u";
	} else {
		while ((logname = get_logname ()) == 0)
			/* do nothing */ ;
	}

	strncpy (logcmd, loginprog, NAME_MAX);
	strncat (logcmd, " ", NAME_MAX - strlen(logcmd));
	strncat (logcmd, logoptstr, NAME_MAX - strlen(logcmd));

	checkname (logname);
	mkarray (logarr, logcmd);
	replacename (logarr, logname);

	if (ch_root)
		chroot (ch_root);
	if (ch_dir)
		chdir (ch_dir);
	if (priority)
		nice (priority);

	sigaction (SIGQUIT, &sa_quit, NULL);
	sigaction (SIGINT, &sa_int, NULL);

	execv (loginprog, logarr);
	error ("%s: can't exec %s: %s", tty, loginprog, strerror (errno));
	sleep (5);
	exit (EXIT_FAILURE);
}
