#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <sys/kd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include "revoke.h"
static int sigfholder(const char *path, int sig);

#ifdef TEST
#define error(fmt,args...)  syslog(LOG_ERR,fmt,args);exit(1)
#define warn(fmt,args...)   syslog(LOG_WARN,fmt,args)
#else
extern void error (const char *fmt, ...);
extern void warn  (const char *fmt, ...);
#endif

#ifndef IUTF8
#  error IUTF8 input flag not defined
#endif

/*
 * Hang up any vcs terminal line
 */
void vcshangup(const gid_t gid, const char * const vcline)
{
	struct stat st;
	char buf[PATH_MAX+1];

	if (!vcline)
		goto skip;

	/* Set up new virtual console text capture devices */
	strncpy (buf, "/dev/vcs", PATH_MAX);
	strncat (buf, vcline, PATH_MAX - strlen(buf));

	if (stat(buf, &st) < 0) {
		if (errno != ENOENT)
			error ("%s: %s", buf, strerror(errno));
		goto skip;
	}
	if (chown (buf, 0, gid) || chmod (buf, (gid ? 0660 : 0600))) {
		if (errno == EROFS)
			warn("%s: %s", buf, strerror(errno));
		else
			error ("%s: %s", buf, strerror(errno));
	}
# ifdef __NR_revoke
	_syscall(int, revoke, const char*, buf);
# else
	if (sigfholder(buf, SIGHUP)) {		/* Be nice, send SIGHUP */
		sleep (2);
		if (sigfholder(buf, SIGKILL))	/* Hey, go away */
			sigfholder(buf, SIGCONT);
	}
# endif /* __NR_revoke */

	/* Set up new virtual console text/attribute capture devices */
	strncpy (buf, "/dev/vcsa", PATH_MAX);
	strncat (buf, vcline, PATH_MAX - strlen(buf));

	if (stat(buf, &st) < 0) {
		if (errno != ENOENT)
			error ("%s: %s", buf, strerror(errno));
		goto skip;
	}
	if (chown (buf, 0, gid) || chmod (buf, (gid ? 0660 : 0600))) {
		if (errno == EROFS)
			warn("%s: %s", buf, strerror(errno));
		else
			error ("%s: %s", buf, strerror(errno));
	}
# ifdef __NR_revoke
	_syscall(int, revoke, const char*, buf);
# else
	if (sigfholder(buf, SIGHUP)) {		/* Be nice, send SIGHUP */
		sleep (2);
		if (sigfholder(buf, SIGKILL))	/* Hey, go away */
			sigfholder(buf, SIGCONT);
	}
# endif /* __NR_revoke */
skip:
	return;
}

/*
 * Reset terminal on stdin
 */
void reset_stdin(const int mode)
{
	struct termios termios;
	/* The above reset only puts the output things into a sane state.
	 * The input state is not reset.
	 */
	memset (&termios, 0, sizeof termios);
	if (tcgetattr (0, &termios))
		warn ("tcgetattr problem: %s", strerror(errno));

	/* Use defaults of <sys/ttydefaults.h> for base settings */
	termios.c_iflag |= TTYDEF_IFLAG;
	termios.c_oflag |= TTYDEF_OFLAG;
	termios.c_lflag |= TTYDEF_LFLAG;
	termios.c_cflag |= (TTYDEF_SPEED | TTYDEF_CFLAG);

	/* Sane setting, allow eight bit characters, no carriage return delay
	 * the same result as `stty sane cr0 pass8'
	 */
	termios.c_iflag |=  (BRKINT | ICRNL | IMAXBEL);
	termios.c_iflag &= ~(IGNBRK | INLCR | IGNCR | IXOFF | IUCLC | IXANY | ISTRIP);
	termios.c_oflag |=  (OPOST | ONLCR | NL0 | CR0 | TAB0 | BS0 | VT0 | FF0);
	termios.c_oflag &= ~(OLCUC | OCRNL | ONOCR | ONLRET | OFILL | OFDEL |\
			     NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY);
	termios.c_lflag |=  (ISIG | ICANON | IEXTEN | ECHO|ECHOE|ECHOK|ECHOKE);
	termios.c_lflag &= ~(ECHONL | NOFLSH | XCASE | TOSTOP | ECHOCTL|ECHOPRT);
	termios.c_cflag |=  (CREAD | CS8 | B38400);
	termios.c_cflag &= ~(PARENB);
#ifdef IUTF8
	/* Set UTF-8 input flag */
	switch(mode) {
		case K_UNICODE:
			termios.c_iflag |= IUTF8;
			break;
		case K_RAW:
		case K_MEDIUMRAW:
		case K_XLATE:
		default:
			termios.c_iflag &= ~IUTF8;
			break;
	}
#endif
	/* VTIME and VMIN can overlap with VEOF and VEOL since they are
	 * only used for non-canonical mode. We just set the at the
	 * beginning, so nothing bad should happen.
	 */
	termios.c_cc[VTIME]    = 0;
	termios.c_cc[VMIN]     = 1;
	termios.c_cc[VINTR]    = CINTR;
	termios.c_cc[VQUIT]    = CQUIT;
	termios.c_cc[VERASE]   = CERASE; /* ASCII DEL (0177) */
	termios.c_cc[VKILL]    = CKILL;
	termios.c_cc[VEOF]     = CEOF;
	termios.c_cc[VSWTC]    = _POSIX_VDISABLE;
	termios.c_cc[VSTART]   = CSTART;
	termios.c_cc[VSTOP]    = CSTOP;
	termios.c_cc[VSUSP]    = CSUSP;
	termios.c_cc[VEOL]     = _POSIX_VDISABLE;
	termios.c_cc[VREPRINT] = CREPRINT;
	termios.c_cc[VDISCARD] = CDISCARD;
	termios.c_cc[VWERASE]  = CWERASE;
	termios.c_cc[VLNEXT]   = CLNEXT;
	termios.c_cc[VEOL2]    = _POSIX_VDISABLE;
	if (tcsetattr (0, TCSADRAIN, &termios))
		warn ("tcsetattr problem: %s", strerror(errno));
}

/*
 * Don't allow the user to pass an option as a user name
 * To be more safe: Use -- to make sure the rest is interpreted
 * as non-options by the program, if it supports it
 */
void checkname (const char* nm)
{
	const char *p = nm;
	if (!nm)
		goto err;
	if (strlen (nm) > 42)
		goto err;
	while (isspace (*p))
		p++;
	if (*p == '-')
		goto err;
	return;
err:
	error ("checkname: %s", strerror(EPERM));
}

void replacename (char** arr, const char* nm)
{
	const char *p;
	char *tmp;
	while ((p = *arr)) {
		const char *p1 = p;
		while (*p1) {
			if (memcmp (p1, "\\u", 2) == 0) {
				tmp = malloc (strlen (p) + strlen (nm));
				if (!tmp)
					error ("replacename: %s", strerror(errno));
				if (p1 != p)
					memcpy (tmp, p, (p1-p));
				*(tmp + (p1-p)) = 0;
				strcat (tmp, nm);
				strcat (tmp, p1+2);
				*arr = tmp;
			}
			p1++;
		}
		arr++;
}
}

void mkarray (char** arr, char* str)
{
	char* p = str;
	char* start = p;
	int i = 0;

	while (*p && i < (ARRAY_SIZE_MAX - 2)) {
		if (isspace (*p)) {
			*p = 0;
			while (isspace (*++p))
				;
			if (*p) {
				arr[i++] = start;
				start = p;
			}
		} else
			p++;
	}
	arr[i++] = start;
	arr[i++] = (char*)0;
}

static int sigfholder(const char *path, int sig)
{
    char buf[MAXNAMLEN+PATH_MAX], *ptr;
    pid_t pid;
    DIR *proc, *fds;
    struct dirent *task, *fd;
    FILE *fp;
    struct stat st, lnk;
    int cnt = 0;

    if ((proc = opendir("/proc")) == (DIR *)0) {
	error("cannot opendir(/proc): %s\n", strerror(errno));
    }

    if (stat(path, &st) < 0) {
	error("cannot stat %s: %s\n", path, strerror(errno));
    }

    while((task = readdir(proc)) != (struct dirent *)0) {
	pid = (pid_t)atoi(task->d_name);

	if (pid == 0 || pid == getpid() || pid == getppid())
	    continue;

	(void) strcpy(buf,"/proc/");
	(void) strcat(strcat(buf, task->d_name), "/maps");
	if ((fp = fopen(buf,"r")) == (FILE*)0) {
	    if (errno != ENOENT)
		warn("cannot open(%s): %s\n", buf, strerror(errno));
	    continue;
	}
	while ((fgets(buf, sizeof(buf)-1, fp))) {
	    if ((ptr = strchr(buf, '/')) == (char*)0)
		continue;
	    if (strncmp(path, ptr, strlen(ptr)-1) == 0) {
		kill(pid, sig);
		cnt++;
		break;
	    }
	}
	fclose(fp);

	(void) strcpy(buf,"/proc/");
        (void) strcat(strcat(buf, task->d_name), "/fd");
	if ((fds = opendir(buf)) == (DIR *)0) {
	    if (errno != ENOENT && errno != EPERM)
		warn("cannot opendir(%s): %s\n", buf, strerror(errno));
	    continue;
	}
	ptr = strcat(buf, "/");
	ptr = ptr + strlen(buf);
	while ((fd = readdir(fds)) != (struct dirent *)0) {
	    if (*fd->d_name == '.')
		continue;
	    (void) strcpy(ptr, fd->d_name);
	    if (stat(buf, &lnk) < 0) {
		continue;
	    }
	    if (st.st_dev == lnk.st_dev && st.st_ino == lnk.st_ino) {
		kill(pid, sig);
		cnt++;
		break;
	    }
	}
	closedir(fds);
    }
    closedir(proc);
    return cnt;
}

#ifdef TEST
int main (int argc, char * argv[])
{
    if (++argv && *argv && (*argv)[0] != '\0')
	printf("%d\n", sigfholder(*argv, SIGHUP));
    return 0;
}
#endif
